@@ -1,16 +1,38 @@
 /*************************************************************************
  * Buddy Timezone Module
- * by Martijn van Oosterhout <kleptog@svana.org> (C) April 2006
- * Licenced under the GNU General Public Licence version 2.
  *
  * A Gaim plugin that allows you to configure a timezone on a per-contact
  * basis so it can display the localtime of your contact when a conversation
  * starts. Convenient if you deal with contacts from many parts of the
  * world.
+ *
+ * by Martijn van Oosterhout <kleptog@svana.org> (C) April 2006
+ * Some code copyright (C) 2006, Richard Laager <rlaager@users.sf.net>
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  *************************************************************************/
 
 #define GAIM_PLUGINS
+
+#ifdef CUSTOM_GTK
+#define PLUGIN "gtk-kleptog-buddytimezone"
+#else
 #define PLUGIN "core-kleptog-buddytimezone"
+#endif
+
 #define SETTING_NAME "timezone"
 #define CONTROL_NAME PLUGIN "-" SETTING_NAME
 
@@ -18,6 +40,7 @@
 #include <errno.h>
 #include <ctype.h>
 #include <string.h>
+#include <math.h>
 
 #include "notify.h"
 #include "plugin.h"
@@ -33,12 +56,23 @@
 
 void *gaim_gtk_blist_get_handle();
 
+#if defined(CUSTOM_GTK) && (GAIM_MAJOR_VERSION < 2)
+//#error Custom GTK Widget only works in Gaim 2
+#undef CUSTOM_GTK
+#endif
+
 //#include "gtkblist.h"   /* gaim_gtk_blist_get_handle */  Requires gtk-dev
 
 #define TIMEZONE_FLAG  ((void*)1)
 #define DISABLED_FLAG ((void*)2)
 
+#define TIME_FORMAT "%X"
+/* Another possible format (hides seconds) */
+//#define TIME_FORMAT  "%H:%M"
+
 static GaimPlugin *plugin_self;
+void *make_timezone_menu(const char *selected);
+const char *get_timezone_menu_selection(void *widget);
 
 /* Resolve specifies what the return value should mean:
  *
@@ -94,41 +128,83 @@ buddy_get_timezone(GaimBlistNode * node, gboolean resolve)
     return timezone;
 }
 
+/* Calcuates the difference between two struct tm's. */
+static float
+timezone_calc_difference(struct tm *remote_tm, struct tm *local_tm)
+{
+    int datediff = 0, diff;
+
+    /* Tries to calculate the difference. Note this only works because the
+     * times are with 24 hours of eachother! */
+    if(remote_tm->tm_mday != local_tm->tm_mday)
+    {
+        datediff = remote_tm->tm_year - local_tm->tm_year;
+        if(datediff == 0)
+            datediff = remote_tm->tm_mon - local_tm->tm_mon;
+        if(datediff == 0)
+            datediff = remote_tm->tm_mday - local_tm->tm_mday;
+    }
+    diff =
+        datediff * 24 * 60 + (remote_tm->tm_hour - local_tm->tm_hour) * 60 + (remote_tm->tm_min -
+                                                                              local_tm->tm_min);
+
+    return (float)diff / 60.0;
+}
+
 #ifdef PRIVATE_TZLIB
 static int
-timezone_get_time(const char *timezone, struct tm *tm)
+timezone_get_time(const char *timezone, struct tm *tm, float *diff)
 {
     struct state *tzinfo = timezone_load(timezone);
+    struct tm *local_tm;
+    time_t now;
+
     if(!tzinfo)
         return -1;
-    else
-    {
-        time_t now = time(NULL);
-        localsub(&now, 0, tm, tzinfo);
-        free(tzinfo);
-    }
+
+    time(&now);
+    localsub(&now, 0, tm, tzinfo);
+    free(tzinfo);
+
+    /* Calculate user's localtime, and compare. If same, no output */
+    local_tm = localtime(&now);
+
+    if(local_tm->tm_hour == tm->tm_hour && local_tm->tm_min == tm->tm_min)
+        return 1;
+
+    *diff = timezone_calc_difference(tm, local_tm);
     return 0;
 }
 #else
 static int
-timezone_get_time(const char *timezone, struct tm *tm_in)
+timezone_get_time(const char *timezone, struct tm *tm, float *diff)
 {
     time_t now;
-    struct tm *tm;
+    struct tm *tm_tmp;
     const gchar *old_tz;
 
     /* Store the current TZ value. */
     old_tz = g_getenv("TZ");
+
     g_setenv("TZ", timezone, TRUE);
-    now = time(NULL);
-    tm = localtime(&now);
-    memcpy(tm_in, tm, sizeof(*tm));
+
+    time(&now);
+    tm_tmp = localtime(&now);
+    *tm = *tm_tmp;              /* Must copy, localtime uses local buffer */
 
     /* Reset the old TZ value. */
     if(old_tz == NULL)
         g_unsetenv("TZ");
     else
         g_setenv("TZ", old_tz, TRUE);
+
+    /* Calculate user's localtime, and compare. If same, no output */
+    tm_tmp = localtime(&now);
+
+    if(tm_tmp->tm_hour == tm->tm_hour && tm_tmp->tm_min == tm->tm_min)
+        return 1;
+
+    *diff = timezone_calc_difference(tm, tm_tmp);
     return 0;
 }
 #endif
@@ -140,6 +216,7 @@ timezone_createconv_cb(GaimConversation * conv, void *data)
     GaimBuddy *buddy;
     struct tm tm;
     const char *timezone;
+    float diff;
     int ret;
 
 #if GAIM_MAJOR_VERSION < 2
@@ -160,18 +237,19 @@ timezone_createconv_cb(GaimConversation * conv, void *data)
     if(!timezone)
         return;
 
-    ret = timezone_get_time(timezone, &tm);
+    ret = timezone_get_time(timezone, &tm, &diff);
 
     if(ret == 0)
     {
 #if GAIM_MAJOR_VERSION > 1
-        char *text = gaim_date_format_short(tm);
+        const char *text = gaim_time_format(&tm);
 #else
         char text[64];
-        strftime(text, sizeof(text), "%X", &tm);
+        strftime(text, sizeof(text), TIME_FORMAT, &tm);
 #endif
 
-        char *str = g_strdup_printf("Remote Local Time: %s", text);
+        char *str = g_strdup_printf("Remote Local Time: %s (%g hours %s)", text, fabs(diff),
+                                    (diff < 0) ? "behind" : "ahead");
 
         gaim_conversation_write(conv, PLUGIN, str, GAIM_MESSAGE_SYSTEM, time(NULL));
 
@@ -180,7 +258,8 @@ timezone_createconv_cb(GaimConversation * conv, void *data)
 }
 
 #if GAIM_MAJOR_VERSION > 1
-static void buddytimezone_tooltip_cb(GaimBlistNode * node, char **text, gboolean full, void *data);
+static void
+buddytimezone_tooltip_cb(GaimBlistNode * node, char **text, gboolean full, void *data)
 #else
 static void
 buddytimezone_tooltip_cb(GaimBlistNode * node, char **text, void *data)
@@ -189,6 +268,7 @@ buddytimezone_tooltip_cb(GaimBlistNode * node, char **text, void *data)
     char *newtext;
     const char *timezone;
     struct tm tm;
+    float diff;
     int ret;
 
 #if GAIM_MAJOR_VERSION > 1
@@ -201,11 +281,24 @@ buddytimezone_tooltip_cb(GaimBlistNode * node, char **text, void *data)
     if(!timezone)
         return;
 
-    ret = timezone_get_time(timezone, &tm);
+    ret = timezone_get_time(timezone, &tm, &diff);
     if(ret < 0)
         newtext = g_strdup_printf("%s\n<b>Timezone:</b> %s (error)", *text, timezone);
+    else if(ret == 0)
+    {
+#if GAIM_MAJOR_VERSION > 1
+        const char *timetext = gaim_time_format(&tm);
+#else
+        char timetext[64];
+        strftime(timetext, sizeof(timetext), TIME_FORMAT, &tm);
+#endif
+
+        newtext =
+            g_strdup_printf("%s\n<b>Local Time:</b> %s (%g hours %s)", *text, timetext, fabs(diff),
+                            (diff < 0) ? "behind" : "ahead");
+    }
     else
-        newtext = g_strdup_printf("%s\n<b>Local Time:</b> %d:%02d", *text, tm.tm_hour, tm.tm_min);
+        return;
 
     g_free(*text);
     *text = newtext;
@@ -216,8 +309,6 @@ buddytimezone_submitfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
 {
     GaimBlistNode *node;
     GaimRequestField *list;
-    const GList *sellist;
-    void *seldata = NULL;
 
     /* timezone stuff */
     gaim_debug(GAIM_DEBUG_INFO, PLUGIN, "buddytimezone_submitfields_cb(%p,%p)\n", fields, data);
@@ -240,6 +331,15 @@ buddytimezone_submitfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
     }
 
     list = gaim_request_fields_get_field(fields, CONTROL_NAME);
+#ifdef CUSTOM_GTK
+    const char *seldata = get_timezone_menu_selection(list->ui_data);
+    if(seldata == NULL)
+        gaim_blist_node_remove_setting(node, SETTING_NAME);
+    else
+        gaim_blist_node_set_string(node, SETTING_NAME, seldata);
+#else
+    const GList *sellist;
+    void *seldata = NULL;
     sellist = gaim_request_field_list_get_selected(list);
     if(sellist)
         seldata = gaim_request_field_list_get_data(list, sellist->data);
@@ -251,8 +351,10 @@ buddytimezone_submitfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
         gaim_blist_node_set_string(node, SETTING_NAME, "none");
     else
         gaim_blist_node_remove_setting(node, SETTING_NAME);
+#endif
 }
 
+#ifndef CUSTOM_GTK
 static int
 buddy_add_timezone_cb(char *filename, void *data)
 {
@@ -261,6 +363,7 @@ buddy_add_timezone_cb(char *filename, void *data)
         gaim_request_field_list_add(field, filename, TIMEZONE_FLAG);
     return 0;
 }
+#endif
 
 static void
 buddytimezone_createfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
@@ -290,6 +393,15 @@ buddytimezone_createfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
     group = gaim_request_field_group_new(NULL);
     gaim_request_fields_add_group(fields, group);
 
+    timezone = buddy_get_timezone(data, FALSE);
+
+#ifdef CUSTOM_GTK
+    field =
+        gaim_request_field_new(CONTROL_NAME,
+                               is_default ? "Default timezone for group" : "Timezone of contact",
+                               GAIM_REQUEST_FIELD_LIST);
+    field->ui_data = make_timezone_menu(timezone);
+#else
     field =
         gaim_request_field_list_new(CONTROL_NAME,
                                     is_default ? "Default timezone for group" :
@@ -300,7 +412,6 @@ buddytimezone_createfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
 
     recurse_directory("/usr/share/zoneinfo/", buddy_add_timezone_cb, field);
 
-    timezone = buddy_get_timezone(data, FALSE);
     if(timezone)
     {
         if(strcmp(timezone, "none") == 0)
@@ -310,6 +421,7 @@ buddytimezone_createfields_cb(GaimRequestFields * fields, GaimBlistNode * data)
     }
     else
         gaim_request_field_list_add_selected(field, "<Default>");
+#endif
 
     gaim_request_field_group_add_field(group, field);
 }
